/*

  (C) 1998 Marco Falcioni UCLA

  SIMULATED ANNEALING:

  Here we take an initial configuration and we melt it.
  Melting is defined by the parameter P->melting_frac, i.e. 
  the temperature at which the fraction of successful updates is at 
  least that.  This is achieved by doubling a low initial temperature 
  till the frac > melting_frac.
  Once T_max is set by this procedure, we thermalize the configuration
  in order to lose correlation with the initial configuration.
  Now the actual annealing will begin.  The temperature is reduced by
  a constant factor till a minimum temperature is reached.
  

*/

#include <stdio.h>
#include "error.h"

#include "atoms.h"
#include "chain.h"
#include "grid.h"
#include "conf.h"
#include "positions.h"
#include "link.h"
#include "misc.h"
#include "oogl.h"
#include "params.h"
#include "functions.h"
#include "verbose.h"
#include "macros.h"

struct Cell;
struct Positions;
struct Rad_pattern;
struct Sym_group;

#define OOGL

void anneal (struct Params * P, const struct Cell * C,
	     const struct Sym_group * G, struct Rad_pattern * RP,
	     struct Positions * T1)
{
  double fraction = 0.0;
  struct Conf * Min;
  struct Grid * Gr;
  struct Positions * T2 = 0;

#ifdef OOGL
  FILE * a;

  a = safe_open("oogl.hist","w");
#endif

  symmetrize_Positions(T1, G);

  T2 = new_Positions(T1->n_unique, T1->n_symmetry, T1->n_actual, 
                     T1->Nmol, T1->M);
  copy_Positions(T1,T2);

  symmetrize_Positions(T2, G);
  mod_Positions(T2);

  /* test that the initial configuration is not singular */
  merge_positions(T2, P, C);

  if(T2->distinct->length < P->n_actual) {
    push_generic_error("biased_anneal","singular initial configuration");
    return;
  }

  Gr = new_Grid(P, C, 0);

  Min = new_Conf(T1->n_unique);

  /* PHASE I:  melt the structure */

  push_verbose_message("anneal","PHASE I: melt the structure");

#ifdef DEBUG
  printf("# PHASE I: melt the structure\n");
#endif


  while (1) {
    char message[1024];
#ifdef DEBUG
  printf("# Melt: T %f SW %ld Radius %f\n", 
	 P->temperature, P->sw_init, P->radius);
  fflush(stdout);
#endif

    fraction = monte_carlo(P, C, G, RP, T1, T2, Gr, Min);

    sprintf(message, "Melting: T = %f fraction %f", P->temperature, fraction);
    push_verbose_message("anneal", message);

#ifdef OOGL
      OOGLCONFP(a, T1, T2, P, C, G, Gr)
#endif

#ifdef DEBUG
    printf("# Frac %f against %f\n", fraction, P->melting_frac);
    fflush(stdout);
#endif
    
    if(fraction > P->melting_frac) break;
    /* increase the temperature and try again */
    P->temperature *= 1.5;
    P->t_max = P->temperature;
  }


  /* PHASE II:  thermalize the structure */

#ifdef DEBUG
  printf("# PHASE II: thermalize the structure\n");
  fflush(stdout);
#endif

  push_verbose_message("anneal","PHASE II: thermalize the structure");


  while(P->sw_therm > 0) {
    char message[1024];
    set_radius(P, fraction);
    fraction = monte_carlo(P, C, G, RP, T1, T2, Gr, Min);

    /*    set_radius(P, fraction, &old); */
    P->sw_therm --;

    sprintf(message,"Therm: radius %f fraction %f", P->radius, fraction);
    push_verbose_message("anneal",message);

#ifdef OOGL
      OOGLCONFP(a, T1, T2, P, C, G, Gr)
#endif

#ifdef DEBUG
    printf("# Therm: accepted %f of moves rad %f\n", fraction, P->radius);
  fflush(stdout);
#endif
  }

  /* PHASE III:  simulated annealing */

#ifdef DEBUG
  printf("# PHASE III: simulated annealing\n");
  fflush(stdout);
#endif
  push_verbose_message("anneal","PHASE III: annealing the structure");
  /* change radius */


  do {
    char message[1024];
    int i;
#ifdef MIN
    double pos[3];
    struct Link * b;
#endif

    scale_temperature(P);
    set_radius(P, fraction);

#ifdef DEBUG
    printf("# Anneal: T %f SW %ld Radius %f\n",
	   P->temperature, P->sw_init, P->radius);
    fflush(stdout);
#endif
    sprintf(message,"temperature %f",P->temperature);
    push_verbose_message("anneal", message);

    for(i = 0; i < P->sw_anneal; i++) {
      fraction = monte_carlo(P, C, G, RP, T1, T2, Gr, Min);
      sprintf(message,"radius %e fraction %e energies:"
	      " Tot %e RP %e Geom  %e", P->radius, fraction,
	      Min->tot_energy, Min->RP_energy, Min->Pos_energy);
      push_verbose_message("anneal",message);

      /*      fprint_Conf(Min, stdout); */

#ifdef OOGL
      OOGLCONFP(a, T1, T2, P, C, G, Gr)
#endif

    }
#ifdef MIN    
    for(b = T1->unique->head, i = 0; i < T1->n_unique; i++, b = b->next) {
      pos[0] = Min->pos[0][i];
      pos[1] = Min->pos[1][i];
      pos[2] = Min->pos[2][i];

      move_Atom(b->atom, pos);
    }
#endif

#ifdef DEBUG
    printf("# accepted %f of moves radius %f\n", fraction, P->radius);
    fflush(stdout);
#endif
    
  }  while (P->temperature > P->t_min);

#ifdef OOGL
  fclose (a);
#endif

  fprint_Conf(Min,stdout);
  
  if(verbose_stream()) 
    write_unique_Positions(T1, verbose_stream());
      
  delete_Positions(T2);
  delete_Grid(Gr);
  delete_Conf(Min);

}
