/*

  (C) 1998 Marco Falcioni UCLA (falcioni@ucla.edu)

  SIMULATED ANNEALING:

  Here we take an initial configuration and we melt it.
  Melting is defined by the parameter P->melting_frac, i.e. 
  the temperature at which the fraction of successful updates is at 
  least that.  This is achieved by doubling a low initial temperature 
  till the frac > melting_frac.
  Once T_max is set by this procedure, we thermalize the configuration
  in order to lose correlation with the initial configuration.
  Now the actual annealing will begin.  The temperature is reduced by
  a constant factor till a minimum temperature is reached.
  
  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html

*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include "error.h"
#include "constants.h"

#include "atoms.h"
#include "chain.h"
#include "grid.h"
#include "conf.h"
#include "positions.h"
#include "link.h"
#include "misc.h"
#include "oogl.h"
#include "trial.h"
#include "params.h"
#include "functions.h"
#include "rad_pattern.h"
#include "verbose.h"
#include "macros.h"
#include "molecule.h"
#define OOGL

struct Cell;
struct Positions;
struct Rad_pattern;
struct Sym_group;

void biased_anneal (struct Params * P, const struct Cell * C,
	     const struct Sym_group * G, struct Rad_pattern * RP,
	     struct Positions * T1)
{
  double fraction = 1.0;
  struct Conf * Min;
  struct Grid * Gr;
  struct Positions * T2 = 0;

#ifdef OOGL
  FILE * a;

  a = safe_open("hist.oogl","a+");
#endif

  symmetrize_Positions(T1, G);

  T2 = new_Positions(T1->n_unique, T1->n_symmetry, T1->n_actual,T1->Nmol,T1->M);
  clone_Positions(T1,T2);

  symmetrize_Positions(T2, G);
  mod_Positions(T2);
  /* test that the initial configuration is not singular */
  merge_positions(T2, P, C);

  Gr = new_Grid(P, C, 0);

  Min = new_Conf(T1->n_unique);

  Min->tot_energy = max_double;



#ifdef DEBUG
  {
  char message[1024];
  copy_Positions(T1,T2);
  symmetrize_Positions(T2,G);
  mod_Positions(T2);

  merge_positions(T2,P,C);

  clean_Grid(Gr);
  fill_Grid(Gr,T2,C);
  Min->Pos_energy = energy (P, C, T2, Gr, Min);

  if(RP != 0) {
    get_intensity(RP, T2, C);
    scale_Rad_pattern(RP);
    Min->RP_energy = distance_Rad_pattern(RP);
  }
  Min->tot_energy =  P->wt_radpat * Min->RP_energy +  Min->Pos_energy;

  sprintf(message,
          "Initial energies:"
          " Tot %e RP %e Geom %e",
          Min->tot_energy, Min->RP_energy, Min->Pos_energy);
  push_verbose_message("anneal",message);
  }
#endif


      OOGLCONFM(a, T1, T2, P, C, G, Gr);

  /* PHASE I:  melt the structure
     NOTE: 4/7/98
     Biased Monte Carlo accepts a lot of moves, and it is therefore unsuitable 
     for use to determine the melting temperature. The melting temperature is
     defined by a given fraction of accepted moves for a fixed value of the
     radius (the amplitude of proposed moves). Therefore we use simple monte 
     carlo to determine melting.  We fix the radius and the fraction 
     and keep doubling the temperature till the observed fraction is
     equal or larger than the one we want
  */

  if(P->sw_init > 0) {
    push_verbose_message("anneal","PHASE I: melt the structure");
    
    while (1) {
/* MWD Change */
      char message[1024];
      fraction = monte_carlo(P, C, G, RP, T1, T2, Gr, Min);
      sprintf(message, "Melting: T = %f fraction %f radius %f", 
	      P->temperature, fraction, P->radius);
      push_verbose_message("anneal", message);
 
#ifdef OOGL
      OOGLCONFM(a, T1, T2, P, C, G, Gr)
#endif
	
      if(fraction > P->melting_frac) /* the radius is as big as we want it */
	break;

      /* increase the temperature and try again */
      P->temperature *= 1.5;
      P->t_max = P->temperature;
    }
  }

  /* PHASE II:  thermalize the structure */

  push_verbose_message("anneal","PHASE II: thermalize the structure");
  
  while(P->sw_therm > 0) {
/* MWD Change */
    char message[1024];
    int i;
    set_radius(P, fraction);
    for(i = 0; i < T1->Nmol; i++) {
      set_phi(T1->M[i], P, fraction);
    }

    fraction = biased_monte_carlo(P, C, G, RP, T1, T2, Gr, Min);

    P->sw_therm --;

    sprintf(message,"Therm: radius %f fraction %f", P->radius, fraction);
    push_verbose_message("anneal",message);

#ifdef OOGL
      OOGLCONFM(a, T1, T2, P, C, G, Gr)
#endif

  }

  /* PHASE III:  simulated annealing */

  push_verbose_message("anneal","PHASE III: annealing the structure");

  do {
/* MWD Change */
    char message[1024];
    int i;

    scale_temperature(P);
   
    set_radius(P, fraction);
   
    for(i = 0; i < T1->Nmol; i++) {
      set_phi(T1->M[i], P, fraction);
    }

    for(i = 0; i < P->sw_anneal; i++) {

      fraction = biased_monte_carlo(P, C, G, RP, T1, T2, Gr, Min); 

      sprintf(message,
	      "s %d temp %e radius %e fraction %e energies:"
	      " Tot %e RP %e Geom %e", 
	      i, P->temperature, P->radius, fraction,
	      Min->tot_energy, Min->RP_energy, Min->Pos_energy);
      push_verbose_message("anneal",message);

      push_verbose_message("anneal", "positions");
      if(verbose_stream())  {
	write_unique_Positions(T1, verbose_stream()); 
        write_Molecule_atoms(T1->Nmol, T1->M, verbose_stream());
      }
#ifdef OOGL
      OOGLCONFM(a, T1, T2, P, C, G, Gr);
#endif
	
    }
    
  }  while (P->temperature > P->t_min );

#ifdef OOGL
  fclose (a);
#endif
/* MWD Change */
/*
  fprint_Conf(Min,stdout);
  
  if(verbose_stream())  {
    write_unique_Positions(T1, verbose_stream());
    write_Molecule_atoms(T1->Nmol, T1->M, verbose_stream());
  }
  */

  delete_Positions(T2);
  delete_Grid(Gr);
  delete_Conf(Min);

  return;

}
