 /* 

   1998 (C) Marco Falcioni UCLA (falcioni@ucla.edu)

   biased montecarlo

   this routine proposes a new configuration
   then accepts the new configuration according to a Metropolis
   test.


  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>

#include "error.h"
#include "memory.h"

#include "atoms.h"
#include "link.h"
#include "params.h"
#include "positions.h"
#include "atoms.h"
#include "cell.h"
#include "grid.h"
#include "functions.h"
#include "chain.h"
#include "misc.h"
#include "trial.h"
#include "symmetry.h"
#include "rad_pattern.h"
#include "oogl.h"
#include "conf.h"
#include "macros.h"
#include "molecule.h"
#include "vector_functions.h"

#ifndef INLINE
#define INLINE
#endif

INLINE
unsigned pick_trial(double W, double * P, unsigned k);

#define MAX_ENE 1.0e10


#undef MC_ENERGY
#undef OOGL

double biased_monte_carlo (const struct Params * P, const struct Cell * C,
		    const struct Sym_group * G, struct Rad_pattern * RP,
		    struct Positions * T1, struct Positions * T2,
		    struct Grid * Gr, struct Conf * Current) 
{
  unsigned i,k = 5; /* the number of trials */
  double Beta = 1.0 / P->temperature;
  unsigned long iter;
  unsigned accepted = 0;
  double * angle, ** dir;
  double * zero_prob;
  double n_energy = 0.0;
  double n_RP_distance = 0.0;
  double *prob, *o_tot_ene;
  unsigned long Accepted = 0, Tried = 0;
  struct Conf * c, ** Confs;  
  struct Trial ** Deltas;
  struct Trial * Delta;
  struct Positions * T3;

#ifdef MC_ENERGY
  FILE * AA, *BB = 0;
  AA = safe_open("energies","w+");
  if(T1->Nmol) BB = safe_open("molecule","w+");
  run_ID(AA);
  if(BB) run_ID(BB);
#endif

#ifdef OOGL
  FILE * AA;
  AA = safe_open("hist.oogl", "a+");
#endif

#ifndef NOCHECKS
  if(P == 0 || C == 0 || G == 0 || T1 == 0 || T2 == 0 || Gr == 0) {
    push_generic_error("monte_carlo", "(nil) args passed");
    exit_clean();
  }
#endif

  /* scratch positions used only in biased MC. */
 /* MWD Change */
  T3 = new_Positions(T1->n_unique, T1->n_symmetry, T1->n_actual,T1->Nmol,T1->M);
  clone_Positions(T1, T3); 
  /* this is needed to ensure all info is equal 6/98 */

  /* malloc kind things */
  c = new_Conf(T1->n_unique);

  /* why is k a variable? eventually it may be input */
  zero_prob = new_vector_d(k);
  prob = new_vector_d(k);
  o_tot_ene = new_vector_d(k);

  Delta = new_Trial(P->radius);
  
  Deltas = (struct Trial **) Malloc(k * sizeof(struct Trial *));
  Confs = (struct Conf **) Malloc(k * sizeof(struct Conf *));
  dir = (double **) Malloc(k * sizeof(double *));
  angle = (double *) Malloc(k * sizeof(double));

  for(i = 0; i < k; i ++) {
    Deltas[i] = new_Trial(P->radius);
    Confs[i] = new_Conf(T1->n_unique);
    dir[i] = (double *) Malloc(3 * sizeof(double));
  }

  /* bootstrap the loop: current energy is the "old energy" */
  copy_Positions(T1,T2);
  symmetrize_Positions(T2,G);
  mod_Positions(T2);   
  merge_positions(T2,P,C);
  clean_Grid(Gr);
  fill_Grid(Gr,T2,C);

  n_energy = energy (P, C, T2, Gr, c);
  
  if(RP) { /* only if we have a RP */
    get_intensity(RP, T2, C, G);
    scale_Rad_pattern(RP);
    n_RP_distance = distance_Rad_pattern(RP);
    c->RP_energy = n_RP_distance;
  }

  o_tot_ene[0] = P->wt_radpat * n_RP_distance + n_energy;
  c->tot_energy = o_tot_ene[0];

  /* keep track of configuration */
  copy_Conf(Current, c);

  /* MC loop */
  for(iter = 0; iter < P->granularity; iter++) {
    unsigned i_atom, mol = 0;
    double Edelta;

    /* loop over unique atoms. we update one at a time */
/* MWD Change */
    for(i_atom = 0; i_atom < T1->n_unique + T2->Nmol; i_atom++) {
      unsigned actual_atom, pick;
      double n_Rosenbluth = 0.0, o_Rosenbluth = 0.0;
      /* seems a nice way to initialize a min */
      double n_min_ene = DBL_MAX, o_min_ene = DBL_MAX;

      Tried++;

      /* pick unique atom to move at random: do not pick 0th atom if
	 you cannot move it */

      actual_atom = (unsigned int) (Random() * (double) (T2->n_unique +
				    T2->Nmol));
      
#ifdef DEBUG
      printf("actual atom %d\n", actual_atom);
#endif

      /* make k new trials */
      for(i = 0; i < k; i++) {
	n_energy = 0.0;
	n_RP_distance = 0.0;
	Confs[i]->RP_energy = 0.0;
	Confs[i]->tot_energy = 0.0;
	zero_prob[i] = 0.0;
	prob[i] = 0.0;

	do {
	  renew_Trial(Deltas[i]);
	  /* if the space group has one or more directions project them off */
#ifdef POLAR
	  if(G->polar && actual_atom == 0) {
	    double x = Deltas[i]->dx;
	    double y = Deltas[i]->dy;
	    double z = Deltas[i]->dz;
	    
	    Deltas[i]->dx = G->polar_matrix[0][0] * x + 
	      G->polar_matrix[0][1] * y + G->polar_matrix[0][2] * z;
 
	    Deltas[i]->dy = G->polar_matrix[1][0] * x + 
	      G->polar_matrix[1][1] * y + G->polar_matrix[1][2] * z;
	    
	    Deltas[i]->dz = G->polar_matrix[2][0] * x + 
	      G->polar_matrix[2][1] * y + G->polar_matrix[2][2] * z;
	  }
#endif
	  copy_Positions(T1,T2);

	  if(actual_atom < T1->n_unique) {
	    move_atom(T2, actual_atom, Deltas[i], C);   
	  } else {
	    mol = actual_atom - T1->n_unique;   /* define mol for this
						   update */
	    shift_Molecule(T2->M[mol], Deltas[i]);
	    
	    ranor(dir[i]);
	    angle[i] = T2->M[mol]->phi * gasdev();
	    
	    rotate_Molecule(T2->M[mol], angle[i], dir[i]);
	  }

	  symmetrize_Positions(T2,G);
	  mod_Positions(T2);
	  merge_positions(T2, P, C);
	  clean_Grid(Gr);
	  fill_Grid(Gr, T2, C);
	  
	  /* here we are filling all the fields of Confs[i] */
	  n_energy = energy(P, C, T2, Gr, Confs[i]);

	} while (n_energy > MAX_ENE); /* rejection mechanism 
					 for overlapping atoms*/

	if(RP) {
	  get_intensity(RP, T2, C, G);
	  scale_Rad_pattern(RP);
	  n_RP_distance = distance_Rad_pattern(RP);
	}

#ifdef DEBUG
	printf("energy %e  distance %e\n", n_energy, n_RP_distance);

#endif
	zero_prob[i] = 1.0;

	Confs[i]->RP_energy = n_RP_distance;
	Confs[i]->tot_energy =  P->wt_radpat * n_RP_distance + n_energy;

	/* remember the minimum energy */
	n_min_ene = min(n_min_ene, Confs[i]->tot_energy);
	
      }

      /* here we want subtract the minimum energy to 
	 the weights. This makes them smaller */
      for(i = 0; i < k; i++) {
	double scra = Confs[i]->tot_energy - n_min_ene;
	prob[i] = exp(-Beta * scra) * zero_prob[i];
	n_Rosenbluth += prob[i];
#ifdef DEBUG
	printf("CONFS tot %e prob %e\n", 
	       Confs[i]->tot_energy, prob[i]);
	if(isnan(prob[i])) exit_clean();
#endif
      }
 
      /* this insures that we do not pick weird configurations in case
         no proposed move makes sense */
      if(n_Rosenbluth == 0.0) continue;
      
      /* we pick the new position */
      pick = pick_trial(n_Rosenbluth, prob, k);

#ifdef DEBUG
	printf("CONFS pick %d wt %e prob %e\n", 
	       pick, n_Rosenbluth, prob[pick]);
#endif

      /* we move to the new position */
      copy_Positions(T1,T2);

      if(actual_atom < T1->n_unique) {
	move_atom(T2, actual_atom, Deltas[pick], C);   
      } else {
	shift_Molecule(T2->M[mol], Deltas[pick]);
	
	rotate_Molecule(T2->M[mol], angle[pick], dir[pick]);
      }

      symmetrize_Positions(T2,G);
      mod_Positions(T2);

      /* we use T3 to get k-1 trial positions around the new position */ 

      o_min_ene = min(o_min_ene, o_tot_ene[0]);
      zero_prob[0] = 1.0;

      for(i = 1; i < k; i ++) {
	double o_energy = 0.0, o_RP_distance = 0.0;
	zero_prob[i] = 0.0;
	o_tot_ene[i] = 0.0;

	do {
	  double aangle, adir[3];
	  renew_Trial(Delta);
#ifdef POLAR	  
	  if(G->polar && actual_atom == 0) {
	    double x = Delta->dx;
	    double y = Delta->dy;
	    double z = Delta->dz;
	    
	    Delta->dx = G->polar_matrix[0][0] * x + 
	      G->polar_matrix[0][1] * y + G->polar_matrix[0][2] * z;
 
	    Delta->dy = G->polar_matrix[1][0] * x + 
	      G->polar_matrix[1][1] * y + G->polar_matrix[1][2] * z;
	    
	    Delta->dz = G->polar_matrix[2][0] * x + 
	      G->polar_matrix[2][1] * y + G->polar_matrix[2][2] * z;
	  }
#endif
	  /* move T3 of Delta and merge positions */
	  copy_Positions(T2,T3);

	  if(actual_atom < T1->n_unique) {
	    move_atom(T3, actual_atom, Delta, C);   
	  } else {
	    shift_Molecule(T3->M[mol], Delta);
	    
	    ranor(adir);
	    aangle = T3->M[mol]->phi * gasdev();
	    
	    rotate_Molecule(T3->M[mol], aangle, adir);
	  }

	  symmetrize_Positions(T3,G);
	  mod_Positions(T3);
	  merge_positions(T3, P, C);
      	  clean_Grid(Gr);
	  fill_Grid(Gr, T3, C);
	  o_energy = energy(P, C, T3, Gr, c);

	} while (o_energy > MAX_ENE); /* rejection mechanism for 
					 overlapping atoms */
	zero_prob[i] = 1.0;
	
	if(RP) {
	  get_intensity(RP, T3, C, G);
	  scale_Rad_pattern(RP);
	  o_RP_distance = distance_Rad_pattern(RP);
	}

	o_tot_ene[i] = P->wt_radpat * o_RP_distance + o_energy;
	o_min_ene = min(o_min_ene, o_tot_ene[i]);
      
      }

      /* again we subtract the minimum energy to make exponential 
	 weigths smaller */
      for(i = 0; i < k; i ++) {
	double scra = o_tot_ene[i] - o_min_ene;
	o_Rosenbluth += exp(-Beta * scra) * zero_prob[i];
      }

      Edelta = 2.0; /* 2 is greater than one */

      if(o_Rosenbluth != 0.0) { /* should NEVER happen */
	Edelta = n_Rosenbluth / o_Rosenbluth;
	/* correct for the minima subtracted */
	Edelta *= exp(-Beta*(n_min_ene - o_min_ene));
      }

      accepted = 1;

      if (Edelta < 1) {
	double b = Random();
	
	if(b > Edelta) accepted = 0; /* reject criteria */

      }

      if(accepted) {
	/* remember the energy we just accepted 
	 we do not need separate RP info */
	o_tot_ene[0] = Confs[pick]->tot_energy;

	copy_Conf(Current, Confs[pick]);

	if(actual_atom < T1->n_unique) {
	  move_atom(T1, actual_atom, Deltas[pick], C);   
	} else {
	  shift_Molecule(T1->M[mol], Deltas[pick]);
	  
	  rotate_Molecule(T1->M[mol], angle[pick], dir[pick]);
	}
#ifdef OOGL
	OOGLCONFP(AA, T1, T2, P, C, G, Gr)
#endif
#ifdef MC_ENERGY
      fprint_Conf(Confs[pick], AA); 
      fflush(AA);
      if(BB) {
	for(iter = 0; iter < T1->Nmol; iter++) {
	  print_Molecule(T1->M[iter], BB);
	}
	fflush(BB);
      }
#endif
	Accepted++;
      } 
    } /* loop over unique atoms */

    /* write the configurations to disk */
 /* MWD Change */
    mod_all_Molecules(T1->Nmol, T1->M, C);

   

  } /* loop over MC sweeps */


  delete_Positions(T3);

  free_vector(o_tot_ene);
  free_vector(prob);
  free_vector(zero_prob);
  
  for(i = 0; i < k; i ++) {
    delete_Trial(Deltas[i]);
    delete_Conf(Confs[i]);
    Free(dir[i]);
  }
  Free(angle);
  Free(dir);
  Free(Deltas);
  Free(Confs);

#ifdef MC_ENERGY
  fclose(AA);
#endif

#ifdef OOGL
  fclose(AA);
#endif


  delete_Trial(Delta);
  delete_Conf(c);

  return (double) Accepted / (double) Tried;
}

/* 
   adapted from Frenkel's book: 
   picks an element of the array P according to its relative weigth 
   k is there for safety and good measure.
*/

INLINE
unsigned pick_trial(double W, double * P, unsigned k)
{
  unsigned i = 0;
  double pick = Random() * W;
  double cump = P[0];

  while(cump < pick && i < k) {
    i++;
    cump += P[i];
  }

  /* make sure we do not over run the array */

  return (i < k ? i : k - 1);

}
