/*

  (C) 1999 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html






*/

#include <stdio.h>
#include <limits.h>
#include <stdlib.h>
#include <math.h>

#include "interpolation.h"
#include "error.h"
#include "matrices.h"
#include "memory.h"
#include "params.h"
#include "atoms.h"
#include "spline.h"
#include "functions.h"
#include "misc.h"
#include "link.h"
#include "chain.h"

#undef DEBUG

double smooth(double a, double b);

double build_framework(const struct Params * P,
		       struct Chain * Neigh,
		       struct Atom * A, double * energies, int * reject)
{
  int i;
  int blen = 0, nblen = 0;
  double nb_ene = 0.0;
  double min_ene = HUGE_VAL; /* in limits.h */
  double *BEne, *BNEne;
  double **Angles, **AEne;
  double * BD, *NBD;
  double sixth = 1.0/6.0;
  double unbond = 0.0;
  struct Link * N;
  struct Atom ** B, ** NB;
  /* these are related to the measurement routine */
  double M_bonds = 0.0, M_bene = 0.0, M_aene = 0.0,
    M_avene = 0.0, M_nbene = 0.0, Far = 0.0;

  * reject = 0;

#ifndef NOCHECKS
  if(P == 0 || Neigh == 0 || A == 0 || energies == 0) {
    push_generic_error("build_framework","(nil) args passed");
    exit_clean();
  }


  if(Neigh->length == 0) {
     push_generic_error("build_framework","empty neighbor list passed");
     exit_clean();
  }

#endif

  B = (struct Atom **) malloc(Neigh->length * sizeof(struct Atom *));
  NB = (struct Atom **) malloc(Neigh->length * sizeof(struct Atom *));
  BD = new_vector_d(Neigh->length);
  NBD = new_vector_d(Neigh->length);
   
  for(N = Neigh->head; N != 0; N = N->next) {
    if(N->key < P->bond) {
      B[blen] = N->atom;
      BD[blen] = N->key;
#ifdef DEBUG
      printf("BD %f\n",BD[blen]);
#endif
      ++blen;
    } else if (N->key > P->bond) {
      NB[nblen] = N->atom;
      NBD[nblen] = N->key;
      ++nblen;
    }
  }

  BEne = new_vector_d(blen); /* energies associated to actual bonds */
  BNEne = new_vector_d(blen); /* energies associated to non-bonds
				 for atoms close enough to be bonded */
  Angles = new_matrix_d(blen); /* holds all possible angles */
  AEne = new_matrix_d(blen); /* holds the energies associated with them */
 
  for(i = 0; i < blen; i++) {
    int j;
    int k,m;
    double factor, over = 0.5;

    /* here I take into account for overcounted bonds...
       bonds to atoms with the same root count 1, the others are overcounted */

    if(B[i]->root == A->root) {
      over = 1.0;
    }

    BEne[i] = over * interp_spline(P->Bond->X, P->Bond->Y, P->Bond->Y2, 
					 P->Bond->n, BD[i], &k, &m);
    BNEne[i] =  over * 
      interp_spline(P->NoBond->X, P->NoBond->Y, P->NoBond->Y2, 
		    P->NoBond->n, BD[i], &k, &m);

    factor = smooth(BD[i],P->bond);

    for(j = 0; j < blen; j++) {
      double angle;
      double factor2;

      if(i >= j) continue;
     
      angle = find_angle(A, B[i], B[j], BD[i], BD[j]);
      factor2 = smooth(BD[j],P->bond);

      if(angle < 0.0) {
	*reject = 1;
	min_ene = DBL_MAX;
	goto end;
      }

      Angles[j][i] = Angles[i][j] = angle;
      AEne[j][i] = AEne[i][j] = factor * factor2 *
	interp_spline(P->Angle->X, P->Angle->Y, P->Angle->Y2, 
		      P->Angle->n, angle, &k, &m);
#ifdef DEBUG 
      printf("angles %ld %ld %f %f\n", i, j, angle, AEne[i][j]);
#endif
    }
    /* diagonal terms */
    Angles[i][i] = 0.0;
    AEne[i][i] = 0.0;

  }
  
  /* non bonded Atoms */
  for(i = 0; i < nblen; i++) {
    int k,m;
    double over = 0.5;
    
    if(NB[i]->root == A->root) {
      over = 1.0;
    } 

    nb_ene += over * interp_spline(P->NoBond->X, P->NoBond->Y, P->NoBond->Y2, 
			    P->NoBond->n, NBD[i], &k, &m);
  }

  Far = nb_ene;

  /* here we do not need a minimization search: 
     just compute the energy and return */

  if(blen < 4) {
    double bond = 0.0, angle = 0.0, avgang = 0.0;
    int a,b, doavangle = 1;

    /* contribution of bonded atoms */
    for(i = 0; i < blen; i++) {
      bond += BEne[i]; 
      M_bonds += BD[i];
    }

    M_bene = bond;
    M_bonds /= (double) blen;

    switch (blen) {
    case 0:      
      /* angle energy is zero: there are no angles */
      doavangle = 0;
      unbond = 500.0;
      break;

    case 1:
      /* angle energy is zero: there are no angles */
      doavangle = 0;
      unbond = 350.0;
      break;

    case 2:
      M_aene = angle = AEne[0][1];
      avgang = Angles[0][1];

      unbond = 200.0; /* the unbond concept was 
				      challenged by Mike 2/10/98 */
      break;
    case 3:
      M_aene = angle = AEne[0][1] + AEne[0][2] + AEne[1][2];
      avgang = .33333333333333333333333 
	* (Angles[0][1] + Angles[0][2] + Angles[1][2]);

      unbond = 100.0;
      break;
    case 4:
      unbond = 0.0;
      break;
    case 5:
      unbond = 300;
      break;
    default:
      unbond = 1000;
    }

    
    /* energy from average angle histogram */
    if(doavangle) 
      M_avene = interp_spline(P->AvAngle->X, P->AvAngle->Y, 
			      P->AvAngle->Y2, P->AvAngle->n, 
			      avgang, &a, &b);
    
#ifdef DEBUG
    printf("Ablen %ld nb_ene %f bond %f angle %f avangle %f unbond %f\n",
	   blen, nb_ene, bond, angle, M_avene, unbond);
#endif

    min_ene = P->wt_nobond * nb_ene
      + P->wt_bond * bond
      + P->wt_angl * angle
      + P->wt_aang * M_avene
      + P->wt_unbond * unbond;

    goto end;

  }
 
  for(i = 0; i < blen; i++) {
    int j;
    for(j  = 0; j < blen; j++) {
      int k;

      if(i >= j) continue;

      for(k = 0; k < blen; k++) {
	int m;

	if(k >= j || k >= i) continue;

	for(m = 0; m < blen; m++) {
	  double energy = nb_ene;
	  double av_angle = 0.0, nob = 0.0;
	  double C_bonds = 0.0, C_bene = 0.0, C_aene = 0.0,
	    C_avene = 0.0, C_nbene = 0.0;
	  int n;
	  int a,b;

	  if(m >= k || m >= i || m >= j) continue;

	  /* add contribution of close enough 
	     nonbonded atoms (overcounted) */
	  for(n = 0; n < blen; n++) {
	    if(n != i && n != j && n != k && n != m) {
	      nob += BNEne[n];
	    }
	  } 
	  /* contribution of non bonded atoms */
	  
	  energy += P->wt_nobond * nob;
	  C_nbene = nob;

	  /* contribution of bonded atoms (overcounted) */

	  C_bonds = BD[i] + BD[j] + BD[k] + BD[m];
	  C_bene = BEne[i] + BEne[j] + BEne[k] + BEne[m];

	  energy += P->wt_bond * C_bene;

	  /* contribution of angle energy */
#ifdef DEBUG2
	  printf("angle enes: %f %f %f %f %f %f\n", 
		 AEne[i][j], AEne[i][k], AEne[i][m], 
		 AEne[j][k], AEne[j][m], AEne[k][m]);
#endif
	  C_aene = AEne[i][j] + AEne[i][k] + AEne[i][m] + 
	    AEne[j][k] + AEne[j][m] + AEne[k][m];
	  
	  energy += P->wt_angl * C_aene;

	  av_angle = Angles[i][j] + Angles[i][k] + Angles[i][m] + 
	    Angles[j][k] + Angles[j][m] + Angles[k][m];
	  av_angle *= sixth;
	 
	  C_avene = interp_spline(P->AvAngle->X, P->AvAngle->Y, 
				  P->AvAngle->Y2, P->AvAngle->n, 
				  av_angle, &a, &b);

	  energy += P->wt_aang * C_avene;
	  
#ifdef DEBUG2
	  printf("%ld %ld %ld %ld blen %ld nb_ene %f bond %f angle %f avanglee %f avang %f\n",
		 i, j, k, m, blen, nb_ene, C_bene, C_aene, C_avene, av_angle);
#endif 
#ifdef DEBUG
	  printf("%ld %ld %ld %ld %f %f\n", i, j, k, m, energy, nob);
#endif
	  
	  if(energy < min_ene) {
	    M_bonds = C_bonds;
	    M_bene = C_bene;
	    M_aene = C_aene;
	    M_avene = C_avene;
	    M_nbene = C_nbene;
	    min_ene = energy;
	  }	  
	  
	  if(blen == 4) { /* one loop is all we need */
	    min_ene = energy;
	    goto end;
	  }
	}
      }
    }
  }

 end:

  /* keep track of atoms too far to be bonded anyhow */
  M_nbene += Far;

  energies[0] = M_bonds;
  energies[1] = M_bene;
  energies[2] = M_aene;
  energies[3] = M_avene;
  energies[4] = M_nbene;
  energies[5] = (double) blen;

  free_vector(BEne);
  free_vector(BNEne);
  free_vector(BD);
  free_vector(NBD);
  free_matrix_d(Angles, blen);
  free_matrix_d(AEne, blen);
  free(B);
  free(NB);
#ifdef DEBUG
	  printf("decided on %f\n", min_ene);
#endif
  return min_ene + P->wt_unbond * unbond;


}
		       
double smooth(double a, double b)
{
  if(a < b - 1.0) return 1.0;
  return -a + b;
}
