/*

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html



  computes the intensites of diffracted X-rays for the given 
  framework geometry.  The "distance" from this computed pattern to 
  the observed pattern is obtained with a call to
  distance_Rad_pattern
  note that the intensities are scaled before the routine exits with a call
  to scale_Rad_pattern

  TODO:
  shall I reuse memory? How much faster can that be?
  shall I scale the pattern here or outside?
  where do I belong?

*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "error.h"
#include "constants.h"
#include "memory.h"

#include "cell.h"
#include "misc.h"
#include "intensity.h"
#include "spot.h"
#include "positions.h"
#include "rad_pattern.h"
#include "chain.h"
#include "link.h"
#include "atoms.h"
#include "molecule.h"
#include "symmetry.h"
#include "group.h"

#define DEBUG

void get_intensity(struct Rad_pattern * RP,
		   const struct Positions * T,
		   const struct Cell * C,
                   const struct Sym_group * G)
{
  int o,d;
  static double twopi = 2.0 * constPi_;
  static struct Atom B;

#ifndef NOCHECKS
  if(RP == 0 || T == 0 || C == 0) {
    push_generic_error("get_intensity","(nil) args passed");
    exit_clean();
  }
#endif

  if(RP->corr == 0) {
    push_generic_error("get_intensity","call get_correction first");
    exit_clean();
  }

  RP->composites = 1;

  for(o = 0; o < T->Nmol; o++) abc_Molecule(T->M[o], C);

  /* Compute F(h,k,l) */

  if(RP->type == XRAY || RP->type == BOTH) {
    int i;
    struct Link * aa;
    double * fx_real;
    double * fx_imag;

    fx_real = new_vector_d(RP->length);
    fx_imag = new_vector_d(RP->length);

    for( i = 0; i < RP->length; i++) {
      fx_real[i] = 0.0;
      fx_imag[i] = 0.0;
    }
   
    /* loop over distinct atoms */

    for(aa = T->distinct->head; aa != 0; aa = aa->next) {
      int j;
      int uniq = aa->atom->root;

      for(j = 0; j < RP->length; j++) {
	double arg;
	double H = RP->I[j]->h_, K = RP->I[j]->k_, L = RP->I[j]->l_;
	
	arg = (H * aa->atom->ra + K * aa->atom->rb + L * aa->atom->rc);
	
	arg = twopi * fmod(arg, 1.0);
      
	fx_real[j] += RP->xray_corr[j][uniq] * cos(arg);
	if(!C->csym) fx_imag[j] += RP->xray_corr[j][uniq] * sin(arg);
      }
    }

    /* molecule code added 10/99 Marco Falcioni */
    for(o = 0, d = 0; o < T->Nmol; o++) {
      int k;

      for(k = 0; k < T->M[o]->N; d++, k++) {
	int j, symno;
	int uniq = T->n_unique + d;

        for (symno = 0; symno < G->n_el; symno++) {
          apply_Group_el(G->list[symno], T->M[o]->A[k], &B);

	  for(j = 0; j < RP->length; j++) {
            double arg;
	    double H = RP->I[j]->h_, K = RP->I[j]->k_, L = RP->I[j]->l_;
	  
	    arg = (H * B.ra + K * B.rb + L * B.rc);
	  
	    arg = twopi * fmod(arg, 1.0);
      
	    fx_real[j] += RP->xray_corr[j][uniq] * cos(arg);
	    if(!C->csym) fx_imag[j] += RP->xray_corr[j][uniq] * sin(arg);
	  }
	}
      }
    }

    for( i = 0; i < RP->length; i++) {
      RP->I[i]->xray->calc = fx_real[i] * fx_real[i];
      if(!C->csym) RP->I[i]->xray->calc += fx_imag[i] * fx_imag[i];
    }

    free_vector(fx_real);
    free_vector(fx_imag);
  }
  
  if(RP->type == NEUTRON || RP->type == BOTH) { 
    int i;
    struct Link * aa;
    double * fn_real;
    double * fn_imag;

    fn_real = new_vector_d(RP->length);
    fn_imag = new_vector_d(RP->length);

    for( i = 0; i < RP->length; i++) {
      fn_real[i] = 0.0;
      fn_imag[i] = 0.0;
    }
    
    /* loop over distinct atoms */
    for(aa = T->distinct->head; aa != 0; aa = aa->next) {
      int j;
      int uniq = aa->atom->root;

      for(j = 0; j < RP->length; j++) {
	double arg;
	double H = RP->I[j]->h_, K = RP->I[j]->k_, L = RP->I[j]->l_;
	
	arg = (H * aa->atom->ra + K * aa->atom->rb + L * aa->atom->rc);

	arg = twopi * fmod(arg, 1.0);
	
	fn_real[j] += RP->neu_corr[j][uniq] * cos(arg);
	if(!C->csym) fn_imag[j] += RP->neu_corr[j][uniq] * sin(arg);
      }
    }


    /* molecule code added 10/99 MWD */
    for(o = 0, d = 0; o < T->Nmol; o++) {
      int k;

      for(k = 0; k < T->M[o]->N; d++, k++) {
        int j, symno;
        int uniq = T->n_unique + d;

        for (symno = 0; symno < G->n_el; symno++) {
          apply_Group_el(G->list[0], T->M[o]->A[k], &B);

          for(j = 0; j < RP->length; j++) {
            double arg;
            double H = RP->I[j]->h_, K = RP->I[j]->k_, L = RP->I[j]->l_;

            arg = (H * B.ra + K * B.rb + L * B.rc);

            arg = twopi * fmod(arg, 1.0);

	    fn_real[j] += RP->neu_corr[j][uniq] * cos(arg);
            if(!C->csym) fn_imag[j] += RP->neu_corr[j][uniq] * sin(arg);
          }
        }
      }
    }

    /* MWD Change */
    for( i = 0; i < RP->length; i++) {
      RP->I[i]->neutron->calc = fn_real[i]*fn_real[i];
      if(!C->csym) RP->I[i]->neutron->calc += fn_imag[i] * fn_imag[i];
    }

    free_vector(fn_real);
    free_vector(fn_imag);
  }

  sum_composites_Rad_pattern(RP);

  return;
}

