/* Created: [GJ] Thu Dec 14 17:06:09 EST 1995
 *
 * [GJ] Mon Feb 10 07:08:25 EST 1997
 * Cleaned up prototypes. Exposed the
 * polynomial and rational function
 * interpolation functions.
 */
#ifndef INTERPOLATION_H_
#define INTERPOLATION_H_


/* Interpolate in a table of (x,y) data using
 * polynomial interpolation, with a polynomial
 * of order n-1, where n is the number of data
 * points. Neville algorithm.
 *
 *  ex:
 *      double X[100], Y[100];
 *      ... calc X, Y...
 *      interp_poly(X, Y, 100, x, &y, &dy);
 *	 ... use y ...
 *
 * An estimate of the error in the interpolation
 * is returned in *dy.
 */
extern void interp_poly(const double * X, const double * Y, int size,
			double x, double * y, double * dy);


/* Interpolate in a table of (x,y) data using
 * rational function interpolation.
 *  
 *  ex:
 *      double X[100], Y[100];
 *      ... calc X, Y...
 *      interp_rat(X, Y, 100, x, &y, &dy);
 *	... use y ...
 *
 * An estimate of the error in the interpolation
 * is returned in *dy.
 *
 * Note: there is something fishy about the implementation
 * of this, which is based on the Num.Rec. code. I believe
 * there is an unavoidable loss of accuracy when the y[]
 * data values are either much larger or much smaller
 * than 1, so that the resulting precision is not
 * invariant under y[] rescaling.
 * Therefore: SCALE y[] DATA TO BE OF ORDER 1 BEFORE USING!
 *
 * This should eventually be replaced by a sane
 * rational interpolation algorithm.
 */
void interp_rat(const double * X, const double * Y, int size,
		double x, double * y, double *dy);


/* Given a tabulated set of function values, construct the
 * "second derivative" array, which is used for subsequent
 * interpolations, implemented by calls to splint().
 *
 * Uses "natural" splines, which set the second derivatives
 * to zero on the boundaries.
 *
 * n = number of points
 *
 * Client is reponsible for freeing the result of,
 * new_spline_y2(), using free_vector().
 */
extern void spline(const double * X, const double * Y, int size, double * y2);
extern double * new_spline_y2(const double * X, const double * Y, int size);


/* Given a tabulated set of function values, construct the
 * "second derivative" array, which is used for subsequent
 * interpolations, implemented by calls to splint().
 *
 * Uses "non-natural" splines, which set the first derivatives
 * to values on the boundaries determined by the points themselves.
 *
 * n = number of points
 *
 * Client is reponsible for freeing the result of
 * new_spline_nn_y2(), using free_vector().
 */
extern void spline_nn(const double * X, const double * Y, int size,
		      double * y2);
extern double * new_spline_nn_y2(const double * X, const double * Y, int n);


/* Evaluate an interpolation given an appropriate
 * "second derivative" array y2[], and the data
 * arrays xa[], ya[].
 *
 * n  = number of points
 * x  = point you want the interpolation at
 * *y = interpolated value
 *
 * Also returns the boundary points for the interval that it finds,
 * as indices into the xa[] array. This is useful if several
 * interpolated evaluations are required at the same value of x.
 * Use with interp_spline_quick() below.
 */
extern double interp_spline(const double * X, const double * Y,
			    const double * y2, int size,
			    double x,
			    int * klo, int * khi
			    );

/* Evaluate an interpolation using a given interval specified by
 * klo and khi. This is just splint() w/o the bisection search. The
 * point is to be able to reuse the results of that search when
 * possible.
 * No check of the values at xa[khi] and xa[klo] is done since it
 * is assumed to have been done when klo and khi were generated.
 */
extern double interp_spline_quick(const double * X, const double * Y,
				  const double * y2,
				  int klo, int khi,
				  double x
				  );


/* Evaluate an interpolation given an appropriate
 * "second derivative" array y2[], the arrays of
 * points xa[], ya[], and two indices into xa[] which represent
 * the endpoints of an x cache.
 * So xa[klocache]---xa[khicache] is a subarray of xa[].
 * This cache array is searched first for an appropriate interval.
 * When the cache misses, a bisection search is done on all of xa[]
 * and a cache of the same size (possibly smaller due to edge effects)
 * is rebuilt.
 *
 * Setting klocache=khicache is equivalent to a cache miss and is
 * a convenient way to reset the cache.
 *
 * n  = number of points
 * x  = point you want the interpolation at
 */
extern double interp_spline_cached(const double * X, const double * Y,
				   const double * y2,
				   int size,
				   double x,
				   int * klocache, int * khicache
				   );


/* Differentiate explicit x,y vector pair by fitting a cubic spline.
 * Assumes y_prime is of sufficient size.
 */
extern void spline_deriv(const double * x, const double * y, int n,
			 double * y_prime);


#endif /* INTERPOLATION_H_ */
