/*

  (C) 1998 Marco Falcioni UCLA


  Parallel tempering.


  Input: file "temperatures"
         format:
	 %d   # number n of temperatures
	 %lf %lf %lf \
	 %lf %lf %lf |
                     |
	 ... ... ... >  n temperatures move amplitudes (radii) and probs
                     |
         %lf %lf %lf |
	 %lf %lf %lf /
	 %lf  # ratio of updates to temperature swap attempts    
	 

	 

  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html




*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "error.h"

#include "atoms.h"
#include "chain.h"
#include "grid.h"
#include "conf.h"
#include "positions.h"
#include "link.h"
#include "misc.h"
#include "oogl.h"
#include "trial.h"
#include "params.h"
#include "functions.h"
#include "rad_pattern.h"
#include "verbose.h"
#include "macros.h"

#define ENERGY

struct Cell;
struct Positions;
struct Rad_pattern;
struct Sym_group;

int test_swap(double, double, double);

double adjust_radius(double, double, double, double, double, double);

void parallel_tempering (struct Params * P, const struct Cell * C,
	     const struct Sym_group * G, struct Rad_pattern * RP,
	     struct Positions * T1)
{
  struct Conf ** Current;
  struct Grid * Gr;
  struct Positions ** TS = 0;
  /* 
     betas holds the tempering temperatures 
     radii holds the amplitude for move proposal
     fractions is quite useless
     swap1 is the swap frequency with N+1
     swap2 is the swap frequency with N+2
  */
  double swap_min, swap_max, frac_min, frac_max;
  double radius_min, radius_max;
  double * betas, * radii, * fractions, *fraction_a;
  double * swap1_a, * swap2_a;
  double * swap1, * swap2, *fix;
  unsigned n_betas, i;
  FILE * Tmp_stream;
  char line[1024], *p;
  double update_ratio;  /* the ratio at which we update vs. swap the structures */

#ifdef OOGL
  FILE ** oogl_streams;
#endif
#ifdef ENERGY
  FILE ** ene_streams;
#endif

  /* set up the parallel tempering scheme */

  Tmp_stream = safe_open("temperatures", "r");

  /* read in the first line: number of temperatures */
  p = get_line(line, sizeof(line), Tmp_stream);
  
  if(sscanf(p, "%d", &n_betas) > 1 || n_betas < 3) {
    push_generic_error("parallel_tempering", 
		       "the temperatures file is garbled! A");
    exit_clean();
  }
  push_verbose_message("parallel_tempering","reading temperatures file");  
  /* arrays */
 
  betas = (double *) Malloc(n_betas * sizeof(double));
  fix = (double *) Malloc(n_betas * sizeof(double));
  radii = (double *) Malloc(n_betas * sizeof(double));
  fractions = (double *) Malloc(n_betas * sizeof(double));
  fraction_a = (double *) Malloc(n_betas * sizeof(double));
  swap1 = (double *) Malloc(n_betas * sizeof(double));
  swap2 = (double *) Malloc(n_betas * sizeof(double));
  swap1_a = (double *) Malloc(n_betas * sizeof(double));
  swap2_a = (double *) Malloc(n_betas * sizeof(double));

#ifdef OOGL
  oogl_streams = (FILE **) Malloc(n_betas * sizeof(FILE *));
#endif
#ifdef ENERGY
  ene_streams = (FILE **) Malloc(n_betas * sizeof(FILE *));
#endif

  TS = (struct Positions **) Malloc(n_betas * sizeof(struct Positions *));
  Current = (struct Conf **) Malloc(n_betas * sizeof(struct Conf *));

  symmetrize_Positions(T1, G);

  p = get_line(line, sizeof(line), Tmp_stream);
  if(sscanf(line, "%lf %lf", &swap_min, &swap_max) != 2) {
    push_generic_error("parallel_tempering",
		       "error reading temperature file B");
    exit_clean();
  }
  p = get_line(line, sizeof(line), Tmp_stream);
  if(sscanf(line, "%lf %lf", &frac_min, &frac_max) != 2) {
    push_generic_error("parallel_tempering",
		       "error reading temperature file B");
    exit_clean();
  } 
  p = get_line(line, sizeof(line), Tmp_stream);
  if(sscanf(line, "%lf %lf", &radius_min, &radius_max) != 2) {
    push_generic_error("parallel_tempering",
		       "error reading temperature file B");
    exit_clean();
  }


  for(i = 0; i < n_betas; i++) {
    char name[1024];
#ifdef OOGL
    sprintf(name, "hist_%d.oogl",i);
    oogl_streams[i] = safe_open(name, "w+");
#endif
#ifdef ENERGY
    sprintf(name, "hist_%d.ene",i);
    ene_streams[i] = safe_open(name, "w+");
#endif

    fix[i] = 0.0;
    swap1[i] = 0.0;
    swap2[i] = 0.0;
    swap1_a[i] = 0.0;
    swap2_a[i] = 0.0;

    fractions[i] = 0.0;
    fraction_a[i] = 0.0;

    /* read first guess */
    p = get_line(line, sizeof(line), Tmp_stream);
    if(sscanf(line, "%lf %lf", 
	      &(betas[i]), &(radii[i])) != 2) {
      push_generic_error("parallel_tempering",
			 "error reading temperature file B");
      exit_clean();
    }

    TS[i] = new_Positions(T1->n_unique, T1->n_symmetry, T1->n_actual);
    clone_Positions(T1,TS[i]);
    /* we need T1 to get the correct atom labels, but then we randomize
       the positons to decorrelate the systems */
    rand_Positions(TS[i]);
    symmetrize_Positions(TS[i], G);
    mod_Positions(TS[i]);
    Current[i] = new_Conf(T1->n_unique);
  }

  fix[0] = 1.0;

  /* read the ratio of updates to swaps */

  p = get_line(line, sizeof(line), Tmp_stream);
  if(sscanf(line, "%lf", &(update_ratio)) != 1 || update_ratio > 1.0 ) {
    push_generic_error("parallel_tempering",
		       "error reading temperature file C");
    exit_clean();
  }

  Gr = new_Grid(P, C, 0);
  
  /*-------------------------------------------------------------------------*
   *           PHASE I:  Thermalize the structure                            *
   *-------------------------------------------------------------------------*/

  /* 
   * Note about configurations:  6/1/98
   * T1 is the configuration we use to bootstrap the parallel tempering
   * routine. Once this is done, it is used as a working copy of the positions
   * to be used inside the biased monte carlo routine.
   *
   */


  push_verbose_message("parallel_tempering",
		       "PHASE I: thermalize the structures");
  
  for(i = 0; i < n_betas; i++) {
    unsigned j;

    for(j = 0; j < P->sw_therm; ++j) {
      /* ugly workaround */
      P->radius = radii[i];
      P->temperature = betas[i];
      fraction_a[i] += 1.0;
      fractions[i] += biased_monte_carlo(P, C, G, RP, TS[i], T1, Gr, Current[i]);
#ifdef ENERGY
      fprintf(ene_streams[i],"%d ", i);
      fprint_Conf(Current[i], ene_streams[i]);
#endif
    }
    
    radii[i] = adjust_radius(radii[i], fractions[i]/fraction_a[i], 
			     frac_max, frac_min, 
			     radius_max, radius_min);

    sprintf(line, "Therm %d: T = %f fraction %f radius %f", i,
	    P->temperature, fractions[i], radii[i]);
    push_verbose_message("parallel_tempering", line);

    fractions[i] = 0.0;
    fraction_a[i] = 0.0;

    for(j = 0; j < P->sw_therm; ++j) {
      /* ugly workaround */
      P->radius = radii[i];
      P->temperature = betas[i];
      fraction_a[i] += 1.0;
      fractions[i] += biased_monte_carlo(P, C, G, RP, TS[i], T1, Gr, Current[i]);
#ifdef ENERGY
      fprintf(ene_streams[i],"%d ", i);
      fprint_Conf(Current[i], ene_streams[i]);
#endif
    }

    radii[i] = adjust_radius(radii[i], fractions[i]/fraction_a[i], 
			     frac_max, frac_min, 
			     radius_max, radius_min);

    sprintf(line, "Therm %d: T = %f fraction %f radius %f", i,
	    betas[i], fractions[i], radii[i]);
    push_verbose_message("parallel_tempering", line);
    
#ifdef OOGL
    OOGLCONFP(oogl_streams[i], TS[i], T1, P, C, G, Gr)
#endif
      
  }  

  /*-------------------------------------------------------------------------*
   *                     PHASE II: Parallel Tempering                        *
   *-------------------------------------------------------------------------*/


  push_verbose_message("parallel_tempering","PHASE II: tempering the structures");
  sprintf(line, "will do %ld PT steps",   P->sw_anneal * n_betas);
  push_verbose_message("parallel_tempering", line);

#define STEPS 200

  for(i = 0 ; i < P->sw_anneal * n_betas; ++i) {
    unsigned j;
    for(j = 0; j < STEPS * n_betas; j++) {
      /* here we decide which action to take: update or swap */
      if(Random() < update_ratio) { /* do an update */
	unsigned pick = (int) ((double) n_betas * Random()); 
	double fraction;

	P->radius = radii[pick];
	P->temperature = betas[pick];
	 
	fraction = 
	  biased_monte_carlo(P, C, G, RP, TS[pick], T1, Gr, Current[pick]);

	fraction_a[pick] += 1.0;

	if(fraction > 0.0) {
	  fractions[pick] += fraction;
	}

#ifdef ENERGY
	fprintf(ene_streams[pick],"%d ", i*STEPS*n_betas + j);
	fprint_Conf(Current[pick], ene_streams[pick]);
#endif
      } else { /* attempt to swap: note we pick fewer cases */
	unsigned pick = floor((n_betas - 1) * Random());
	double ene1 = Current[pick]->tot_energy;
	double ene2 = Current[pick + 1]->tot_energy;
	double new, old, delta;

	swap1_a[pick] += 1.0; /* attempted */
	swap2_a[pick] += 1.0; /* attempted */

	old = ene1 / betas[pick] + ene2 / betas[pick + 1];
	new = ene2 / betas[pick] + ene1 / betas[pick + 1];
	delta = new - old;

	/* mock swap with system two steps away */
	if(pick < n_betas - 2) {
	  double old2, new2, delta2, ene3;
	  ene3 = Current[pick + 2]->tot_energy;
	  old2 = ene1 / betas[pick] + ene3 / betas[pick + 2];
	  new2 = ene3 / betas[pick] + ene1 / betas[pick + 2];
	  delta2 = new2 - old2;
	  if(Random() < exp(-delta2)) {
	    swap2[pick] += 1.0;
	  }
	}

	if(Random() < exp(-delta)) { /* swap configurations */
	  struct Positions * T_tmp = TS[pick];
	  struct Conf * C_tmp = Current[pick];

	  swap1[pick]+= 1.0; /* succesful swap */

	  TS[pick] = TS[pick + 1];
	  Current[pick] = Current[pick + 1];
	  TS[pick + 1] = T_tmp;
	  Current[pick + 1] = C_tmp;
	  
#ifdef ENERGY
	  fprintf(ene_streams[pick], "# step %d: swapped with %d\n%d ", i, 
		  pick + 1, i* STEPS*n_betas + j);
	  fprint_Conf(Current[pick], ene_streams[pick]);
	  fprintf(ene_streams[pick + 1], "# step %d: swapped with %d\n%d ", i, 
		  pick, i* STEPS*n_betas + j);
	  fprint_Conf(Current[pick + 1], ene_streams[pick + 1]);
#endif
	}
      

#ifdef OOGL
      OOGLCONFP(oogl_streams[pick], TS[pick], T1, P, C, G, Gr)
      OOGLCONFP(oogl_streams[pick + 1], TS[pick + 1], T1, P, C, G, Gr)
#endif
	} /* MC move or swap */
    } /* granularity for new PT */

    /* MC parameters */

    for(j = 0; j < n_betas; j++) {
      double F = fractions[j]/fraction_a[j];

      radii[j] = 
	adjust_radius(radii[j], F, frac_max, frac_min,
		      radius_max, radius_min);
      fractions[j] = 0.0;
      fraction_a[j] = 0.0;
    }

    /* swap tests */

    printf("New round\n");

    for(j = 0; j < n_betas - 1; j++) {
      int S1 = 3, S2 = 3;

      if(fix[j+1] > 0.0) continue;

      if(swap1_a[j] != 0.0) {
	S1 = test_swap(swap1[j]/swap1_a[j], swap_max, swap_min);
      }

      if(swap2_a[j] != 0.0) {
	S2 = test_swap(swap2[j]/swap2_a[j], swap_max, swap_min);
      }

      if(S1 == 0 && j != n_betas-1) { /* insert */
	printf("Move down\n");
	/*	for(k = n_betas - 1; k > j + 1; k--) {
	  swap1[k] = swap1[k-1];
	  swap1_a[k] = swap1_a[k-1];
	  swap2[k] = swap2[k-1];
	  swap2_a[k] = swap2_a[k-1];
	  
	  betas[k] = betas[k-1];
	  radii[k] = radii[k-1];
	}

	swap1[j+1] = 0.0;
	swap1_a[j+1] = 0.0;
	swap2[j+1] = 0.0;
	swap2_a[j+1] = 0.0;

	betas[j+1] = sqrt(betas[j+2] * betas[j]); 
	radii[j+1] = sqrt(radii[j+2] * radii[j]);  geometric average */

	betas[j + 1] = 0.85 * betas[j + 1];
	if(betas[j] >= betas[j+1]) betas[j+1] = 1.05 * betas[j];
	
      } else if (S1 >= 1 && S2 > 0 && fix[j+1] < 1.0) { /* kill */
	int k;
	printf("Kill\n");
	for(k = j+1; k < n_betas - 1; k++) {
	  swap1[k] = swap1[k+1];
	  swap1_a[k] = swap1_a[k+1];
	  swap2[k] = swap2[k+1];
	  swap2_a[k] = swap2_a[k+1];
	  
	  betas[k] = betas[k+1];
	  radii[k] = radii[k+1];
	}
	swap1[n_betas - 1] = 0.0;
	swap1_a[n_betas - 1] = 0.0;
	swap2[n_betas - 1] = 0.0;
	swap2_a[n_betas - 1] = 0.0;

	betas[n_betas - 1] = 1.2 * betas[n_betas - 2];
	radii[n_betas - 1] = radii[n_betas - 2];
	
	break;
      } else if (S1 == 2) { /* move up */
	printf("Move Up\n");
	if(j != n_betas - 2) {
	  betas[j + 1] = 1.15 * betas[j + 1];
	} else {
	  if(radii[j+1] < radius_max) 
	    betas[j + 1] = 1.5 * betas[j + 1];
	}
      } else {
	if(fix[j] > 0.0 && fix[j+1] < 1.0) {
	  fix[j+1] = 1.0;
	  printf("%d fixed\n", j+1);  
	}
	/* nop */
      }
    }
    for(j = 0; j < n_betas; ++j){
      printf("%d %f %f %f %f %f %f\n", j, 
	     betas[j], radii[j], swap1[j], swap1_a[j], swap2[j], swap2_a[j]);
    }
    
    fflush(stdout);

    for(j = 0; j < n_betas; ++j){
      swap1[j] = 0.0; 
      swap1_a[j] = 0.0; 
      swap2[j] = 0.0; 
      swap2_a[j] = 0,0;
    }
    
  }

#ifdef OOGL
  for(i = 0; i < n_betas; ++i) {
    fclose (oogl_streams[i]);
  }
  Free(oogl_streams);
#endif
#ifdef ENERGY
  for(i = 0; i < n_betas; ++i) {
    fclose (ene_streams[i]);
  }
  Free(ene_streams);
#endif

  delete_Grid(Gr);

  for(i=0; i < n_betas; i++) {
    delete_Positions(TS[i]);
    delete_Conf(Current[i]);
  }

  Free(TS);
  Free(Current);

  Free(betas);
  Free(radii);
  Free(fractions);
  Free(fraction_a);
  Free(swap1);
  Free(swap1_a);
  Free(swap2);
  Free(swap2_a);

  return;

}

double adjust_radius(double R, double F, double F_max, double F_min, 
		     double R_max, double R_min)
{
  if(F > F_max) {
    R *= 1.1;
    if (R > R_max) R = R_max;
  } else if (F < F_min) {
    R *= 0.9;
    if(R < R_min) R = R_min;
  }
  return R;
}

int test_swap(double S, double S_max, double S_min){
  if(S < S_min) return 0;
  if(S > S_max) return 2;
  return 1;
}




