/*

  (C) 1997 Marco Falcioni UCLA


  this routines reads the parameters in the data file,
  writes the parameters into human readable form.


  (C) 1997 Marco Falcioni (falcioni@ucla.edu)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307,
  USA.

  http://www.fsf.org/copyleft/gpl.html



*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <errno.h>
#include <time.h>

extern int errno;

#include "error.h"

#include "misc.h"
#include "verbose.h"
#include "params.h"

const char DATAPATH[20] = "DATA_PATH";
const int LINELENGTH = 512;
const int BUFSIZE = 256;

struct Params * read_Params(FILE * stream)
{
  struct Params * P;
  size_t len;
  int c, itmp;
  char error[200] = "";
  char * line;
  char * p;
  char * buf;

  if(stream == 0) {
    push_generic_error("read_Params", "(nil) arg passed");
    exit_clean();
  }

  P = (struct Params *) Malloc(sizeof(struct Params));

  if(P == 0) {
    push_error("read_Params", Error_Alloc_);
    exit_clean();
  }

  push_verbose_message("read_params","reading parameters");
  
  line = (char *) Malloc(LINELENGTH * sizeof(char));

  /* max bond length */
  p = get_line(line, LINELENGTH , stream);
  c = sscanf(p, "%lf", &(P->bond));
  if(c != 1) strcat(error, "bond ");
  P->bond2 = P->bond * P->bond;

  /* max range of repulsive force */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf", &(P->frange));
  if(c != 1) strcat(error, "force range ");
  P->frange2 = P->frange * P->frange;

  /* max merging distance */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf %lf", &(P->merge2), &(P->ene_merge));
  if(c != 2) strcat(error, "merge/ene_merge ");
  P->merge2 = P->merge2 * P->merge2;

  /* actual number of atoms (density) */
  p = get_line(line, LINELENGTH, stream);
/* MWD Change */
  c = sscanf(p, "%d", &(itmp));
  if(itmp < 0) {
    push_generic_error("read_Params",
                "Please edit the input file and	specify a density");
    exit_clean();
  }	
  P->n_actual = itmp;
  if(c != 1) strcat(error, "n_actual ");

  /* energy weigths */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf %lf %lf %lf %lf %lf", &(P->wt_bond), 
	     &(P->wt_nobond), &(P->wt_angl), &(P->wt_aang), 
	     &(P->wt_merge), &(P->wt_unbond));
  if(c < 6) strcat(error, "weights ");

  /* weigth for radiation pattern */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf", &(P->wt_radpat));
  if(c != 1) strcat(error, "rad_pat wt ");

  /* random seed */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%ld", &(P->seed));
  if(c != 1) strcat(error, "seed ");
 
  /* Max/Min proposed move amplitude */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf %lf %lf", &(P->r_max), &(P->r_min), &(P->gamma));
  if(c != 3) strcat(error, "R_max gamma ");

  P->radius = P->r_max; /* initial value */
 
  /* SIMULATED ANNEALING 
     starting temperature */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf %lf %ld", &(P->t_0), &(P->t_min), &(P->sw_init));
  if(c != 3) strcat(error, "temperatures ");

  P->temperature = P->t_0; /* initial temperature */
  P->t_max = - 1.0; /* this is unknown */

  /* Thermalization sweeps */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%ld %ld %ld", &(P->sw_therm), &(P->sw_anneal), 
	     &(P->granularity));
  if(c != 3) strcat(error, "Sweeps ");

  /* Melting fraction / MC fraction */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf %lf", &(P->melting_frac), &(P->mc_frac));
  if(c != 2) strcat(error, "melting fraction / MC fraction ");

  /* factor reducing temperature in SA */
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%lf", &(P->alpha));
  if(c != 1) strcat(error, "reduction factor ");

   /* Bond spline data filename */
  buf = (char *) Malloc(BUFSIZE * sizeof(char));
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%s", buf);
  if(c != 1) strcat(error, "bond spline ");
  P->Bond_Spl_file = buf;

  /* NoBond spline data filename */
  buf = (char *) Malloc(BUFSIZE * sizeof(char));
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%s", buf);
  if(c != 1) strcat(error, "no_bond spline ");
  P->NoBond_Spl_file = buf;
  
  /* Angle spline data filename */
  buf = (char *) Malloc(BUFSIZE * sizeof(char));
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%s", buf);
  if(c != 1) strcat(error, "angle spline ");
  P->Angle_Spl_file = buf;

  /* Average Angle spline data filename */
  buf = (char *) Malloc(BUFSIZE * sizeof(char));
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%s", buf);
  if(c != 1) strcat(error, "av_angle spline ");
  P->AvAngle_Spl_file = buf;

  /* structure factor table */
  buf = (char *) Malloc(BUFSIZE * sizeof(char));
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%s", buf);
  if(c != 1) strcat(error, "scat fact file ");
  P->Scat_Fact_file = buf;

  /* known powder data intensity */
  P->Rad_file = 0;
  buf = (char *) Malloc(BUFSIZE * sizeof(char));
  p = get_line(line, LINELENGTH, stream);
  c = sscanf(p, "%s", buf);
  if(c != 1) strcat(error, "radiation file ");
  /* strcmp works backwards */
  if(strcmp(buf, "none")) P->Rad_file = buf;
  
  if(strcmp(error, "")) {
    char buf[1024];
    strcpy(buf,"error reading parameters from stream: ");
    strcat(buf, error);
    push_generic_error("read_params",buf);
    exit_clean();
  }

  /* path for data files 
     behavior: the environment variable overrides the value in the file */
  buf = (char *) Malloc(BUFSIZE * sizeof(char));
  p = getenv(DATAPATH);

  if(p == 0) {
    p = get_line(line, LINELENGTH, stream);
    sscanf(p, "%s", buf);
  } else {
    get_line(line, LINELENGTH, stream); /* gobble line */
    strcpy(buf, p);
  }

  P->data_path = buf;

  /* ensure the path ends with a slash */

  len = strlen(buf);

  if(buf[len-1] != '/')  {
    buf[len] = '/';
    buf[len+1] = '\0';
  } 

  if(len > BUFSIZE) {
    push_generic_error("read_Params",
		       "the data path is too long, check BUFSIZE in params.c");
    exit_clean();
  }

  /* attach path to filenames, only if filenames are not preceded by / */

  if(len > 1 && P->Bond_Spl_file[0] != '/') {
    strcpy(line,buf);
    strcat(line,P->Bond_Spl_file);
    strcpy(P->Bond_Spl_file,line);

    line[0] = '\0';
    strcpy(line,buf);
    strcat(line,P->NoBond_Spl_file);
    strcpy(P->NoBond_Spl_file,line);

    line[0] = '\0';
    strcpy(line,buf);
    strcat(line,P->Angle_Spl_file);
    strcpy(P->Angle_Spl_file,line);

    line[0] = '\0';
    strcpy(line,buf);
    strcat(line,P->AvAngle_Spl_file);
    strcpy(P->AvAngle_Spl_file,line);

    line[0] = '\0';
    strcpy(line,buf);
    strcat(line,P->Scat_Fact_file);
    strcpy(P->Scat_Fact_file,line);
    
  }

  Free(line);

  return P;
}

void delete_Params(struct Params * P)
{
  if(P == 0) return;
  Free(P->Bond_Spl_file);
  Free(P->NoBond_Spl_file);
  Free(P->Angle_Spl_file);
  Free(P->AvAngle_Spl_file); 
  Free(P->Scat_Fact_file);
  Free(P->Rad_file);
  Free(P->data_path);
  Free(P);
}


void write_Params(const struct Params * P, FILE * stream)
{
  char * user;
  char * pwd;
  char * host;
  char * hosttype;
  char * date;
  time_t t;

  if(P == 0 || stream == 0) {
    push_generic_error("print_Params","(nil) args passed");
    exit_clean();
  }

  t = time(NULL);
  date = asctime(localtime(&t));  

  user = getenv("USER");
  pwd = getenv("PWD");
  host = getenv("HOST");
  hosttype = getenv("HOSTTYPE");

  fprintf(stream, "# (C) 1997 Marco Falcioni & Michael W. Deem UCLA\n"); 
  fprintf(stream, "# executed on %s", date); 
  fprintf(stream, "# by %s on %s (%s)\n", user, host, hosttype);
  fprintf(stream, "# the current directory is %s\n", pwd);
  fprintf(stream, "# Parameters used: \n");
  fprintf(stream, "%6.2f \t \t  # max bonding distance (A)\n", P->bond);
  fprintf(stream, "%6.2f \t \t  # max non-bonding distance (A)\n", P->frange);
  fprintf(stream, "%6.2f %8.2f \t  # max merging distance (A) and min energy\n", 
	  sqrt(P->merge2), P->ene_merge);
  fprintf(stream, "%ld \t \t  # number of T atoms (density)\n", P->n_actual);
  fprintf(stream, "%6.2f %6.2f %6.2f %6.2f %6.2f %6.2f #"
	  " energy weigths B NB An AvA Mrg\n", 
	  P->wt_bond, P->wt_nobond, P->wt_angl, 
	  P->wt_aang, P->wt_merge, P->wt_unbond);
  fprintf(stream, "%6.2f \t \t  # Radiation Pattern weigth\n", P->wt_radpat);
  fprintf(stream, "%ld \t \t  # Random seed\n", P->seed);
  fprintf(stream, 
	  "%6.2f %6.2f %6.2f  # move amplitudes and control gamma\n", 
	  P->r_max, P->r_min, P->gamma);
  fprintf(stream, 
	  "%10.2f %10.2f %ld  # Simulated annealing temps T_o T_min Init_flag\n",
	  P->t_0, P->t_min, P->sw_init);
  fprintf(stream, "%ld %ld %ld  # Therm Sweeps / SA steps / granularity\n", 
	  P->sw_therm, P->sw_anneal, P->granularity);
  fprintf(stream, "%6.2f %6.2f \t  # Melting acceptance fraction / mc frac \n", 
	  P->melting_frac, P->mc_frac);
  fprintf(stream, "%6.2f \t \t # Temperature reduction factor\n", P->alpha);
  fprintf(stream, "%s \t # spline filenames: Bond\n", P->Bond_Spl_file);
  fprintf(stream, "%s \t # Nobond\n", P->NoBond_Spl_file);
  fprintf(stream, "%s \t # Angle\n", P->Angle_Spl_file);
  fprintf(stream, "%s \t # AvAngle\n", P->AvAngle_Spl_file);
  fprintf(stream, "%s \t # Scattering data\n", P->Scat_Fact_file);
  if(P->Rad_file != 0) {
    fprintf(stream, "%s \t # Scattering intensities data\n", P->Rad_file);
  }
  else {
    fprintf(stream, "none # Scattering intensities data\n");
  }
  fprintf(stream, "%s # Data path\n", P->data_path);

  if(ferror(stream)) {
    push_generic_error("print_Params","error while writing to stream");
    exit_clean();
  }

  return;
}

int set_radius(struct Params * P, double fraction)
{
#ifndef NOCHECKS
  if(P == 0 || fraction < 0.0 ) {
    push_generic_error("set_radius","(nil) arg passed");
    exit_clean();
  }
#endif
  
  P->radius *= 1 + P->gamma * (fraction - P->mc_frac);

  if(P->radius < P->r_min) {
    P->radius = P->r_min;
    return 1;
  }
  /* this is a safety measure to prevent blowups */
  if(P->radius > P->r_max) {
    P->radius = P->r_max;
    return 1;
  }
  return 0;
}

double scale_temperature(struct Params * P) 
{
  double delta;
  double t_old;
#ifndef NOCHECKS
  if(P == 0) {
    push_generic_error("scale_temperature","(nil) arg passed");
    exit_clean();
  }
#endif
  t_old = P->temperature;
  P->temperature = P->alpha * P->temperature;

  delta = t_old - P->temperature;
  return delta;
}


/*
void main (void) {
  struct Params * P;

  connect_error_stream(stderr);
  P = read_Params(stdin);

  print_Params(P, stdout);

}
*/

